import os

from kubernetes.client.rest import ApiException
import yaml
import json

import si_tests.utils.templates as template_utils
from si_tests import settings
from si_tests.clients.k8s import K8sCluster
from si_tests.clients.k8s.base import K8sNamespacedResource
from si_tests.clients.k8s.pods import K8sPod
from si_tests import logger
from si_tests.utils import utils, waiters

LOG = logger.logger


class TFManager(object):
    kubeconfig = None
    _tfoperator_version = None

    def __init__(self, kubeconfig=settings.KUBECONFIG_PATH):
        self._api = None
        self.kubeconfig = kubeconfig
        self.tf_namespace = "tf"
        self.tfoperator_name = "openstack-tf"

    @property
    def api(self):
        """
        :rtype: cluster.K8sCluster
        """
        if self._api is None:
            if not os.path.isfile(self.kubeconfig):
                raise FileNotFoundError(
                    "kubeconfig file={} not found!".format(self.kubeconfig)
                )
            self._api = K8sCluster(kubeconfig=self.kubeconfig)
        return self._api

    def get_k8s_resource(self, resources_type, resource_name, namespace,
                         read=False):
        resources_api = getattr(self.api, resources_type)
        resource = resources_api.get(name=resource_name, namespace=namespace)
        if read:
            return resource.read()
        return resource

    def get_tfoperator_deployment(self, read=False):
        return self.get_k8s_resource(
            "deployments", "tungstenfabric-operator", self.tf_namespace, read
        )

    def tfoperator(self, read=False):
        return self.get_k8s_resource(
            "tfoperator", self.tfoperator_name, self.tf_namespace, read
        )

    def tfanalytics(self, read=False):
        return self.get_k8s_resource(
            "tfanalytics", "tf-analytics", self.tf_namespace, read
        )

    def tfconfig(self, read=False):
        return self.get_k8s_resource(
            "tfconfig", "tf-config", self.tf_namespace, read
        )

    def tfcontrol(self, read=False):
        return self.get_k8s_resource(
            "tfcontrol", "tf-control", self.tf_namespace, read
        )

    def tfvrouter(self, read=False):
        return self.get_k8s_resource(
            "tfvrouter", "tf-vrouter", self.tf_namespace, read
        )

    def tftool(self, read=False):
        return self.get_k8s_resource(
            "tftool", "tf-tool", self.tf_namespace, read
        )

    def tfwebui(self, read=False):
        return self.get_k8s_resource(
            "tfwebui", "tf-webui", self.tf_namespace, read
        )

    def tfimageprecaching(self, read=False):
        return self.get_k8s_resource(
            "tfimageprecaching", "tf-image-pre-caching", self.tf_namespace, read
        )

    def tfoperator_version(self, cache=False):
        if self._tfoperator_version is None or cache is False:
            tfoperator_deployment = self.get_tfoperator_deployment()
            tf_deployment = tfoperator_deployment.read()
            tf_containers = tf_deployment.spec.template.spec.containers
            tf_container = [c for c in tf_containers if c.name == "manager"][0]
            self._tfoperator_version = tf_container.image.split(':')[-1]

        return self._tfoperator_version

    def get_vrouter_pods(self):
        pods = []
        for kind in ['agent', 'agent-dpdk']:
            label = ('app=tf-vrouter-{}'.format(kind))
            pods_filtered = self.api.pods.list(namespace=self.tf_namespace,
                                               label_selector=label)
            pods.extend(pods_filtered)
        return pods

    def get_analytics_nodes(self):
        nodes = []
        selectors = ['tfanalytics=enabled,tfanalyticsdb=enabled',
                     'tfanalytics=disabled,tfanalyticsdb=disabled']
        for selector in selectors:
            nodes.extend(self.api.nodes.list(
                namespace=self.tf_namespace,
                label_selector=selector
            ))
        return nodes

    def get_tf_services_cfgmap(self):
        return self.api.configmaps.get(name="tf-services-cfgmap",
                                       namespace=self.tf_namespace)

    def _get(self, cr, path, default=None):
        obj = cr
        for key in path.split(':'):
            obj = obj.get(key, {})
        if not obj:
            obj = default
        return obj

    def is_tf_cli_enabled(self):
        return self._get(self.tfoperator().data,
                         'spec:controllers:tf-tool:tf-cli:enabled',
                         default=False)

    def enable_tf_cli(self, state=True):
        LOG.info(f"Set TF cli to ${state}")
        tfoperator = self.tfoperator()
        current = self.is_tf_cli_enabled()
        if current == state:
            LOG.info(f"TF cli state is already: {current}")
        else:
            patch = {
                "spec": {
                    "controllers": {
                        "tf-tool": {
                            "tf-cli": {
                                "enabled": state
                            }
                        }
                    }
                }
            }
            tfoperator.patch(patch)

    def is_ctools_enabled(self):
        return self._get(self.tfoperator().data,
                         'spec:controllers:tf-tool:tools:enabled',
                         default=False)

    def enable_ctools(self, state=True, labels=None):
        LOG.info(f"Set TF contrail-tools to ${state}")
        if not labels:
            labels = {'tfvrouter': 'enabled'}
        tfoperator = self.tfoperator()
        current = self.is_ctools_enabled()
        if current == state:
            LOG.info(f"TF contrail-tools (ctools) state is already: {current}")
        else:
            patch = {
                "spec": {
                    "controllers": {
                        "tf-tool": {
                            "tools": {
                                "enabled": state,
                                "labels": labels
                            }
                        }
                    }
                }
            }
            tfoperator.patch(patch)

    def is_analytics_disabled(self):
        return self._get(self.tfoperator().data,
                         'spec:settings:disableTFAnalytics',
                         default=False)

    def disable_analytics(self, state=False):
        LOG.info(f"Set TF Analytics to ${state}")
        tfoperator = self.tfoperator()
        current = self.is_analytics_disabled()
        if current == state:
            LOG.info(f"disableTFAnalytics state is already: {current}")
        else:
            patch = {
                "spec": {
                    "settings": {
                        "disableTFAnalytics": state
                    }
                }
            }
            tfoperator.patch(patch)
        self.wait_tf_analytics(exist=not state)

    def wait_tf_analytics(self, exist=True, timeout=180, interval=10):
        tfanalytics = self.tfanalytics()
        if exist:
            LOG.info("Wait until TF Analytics resource is created")
            check_func = self._is_k8s_obj_exist
        else:
            LOG.info("Wait until TF Analytics resource is deleted")
            check_func = self._is_k8s_obj_not_exist
        waiters.wait(
            check_func,
            predicate_args=[tfanalytics],
            interval=interval,
            timeout=timeout,
        )

    def run_pytest(self):
        # Run TF verification tests (tungsten-pytest) and return pod status.
        LOG.info("Run TF Test (tungsten-pytest)")
        tftest_pod_name = "tf-test-tungsten-pytest"
        tfoperator = self.tfoperator()
        # Check current tf-test status
        tftest = self._get(
            tfoperator.data,
            'spec:controllers:tf-test:tungsten-pytest:enabled',
            default=False)
        patch = {
            "spec": {
                "controllers": {
                    "tf-test": {
                        "tungsten-pytest": {
                            "containers": [{
                                "name": "tungsten-pytest",
                                "env": [{
                                    "name": "PYTEST_MARKEXPR",
                                    "value": "smoke"
                                }]
                            }],
                            "enabled": False
                        }
                    }
                }
            }
        }
        if tftest:
            LOG.info("Disable tungsten-pytest before new run")
            # Disable TF Test
            tfoperator.patch(patch)
            # Wait until pod will be deleted
            waiters.wait(
                lambda: not self.api.pods.present(
                    tftest_pod_name, namespace=self.tf_namespace),
                timeout=180, interval=10
            )
        LOG.info("Enable tungsten-pytest")
        patch["spec"]["controllers"]["tf-test"]["tungsten-pytest"][
            "enabled"] = True
        tfoperator.patch(patch)
        waiters.wait(
            lambda:
            self.api.pods.present(tftest_pod_name, namespace=self.tf_namespace),
            timeout=180, interval=10
        )
        test_pod = self.api.pods.get(tftest_pod_name,
                                     namespace=self.tf_namespace)
        LOG.info("Wait for pod to be Running...")
        test_pod.wait_phase('Running')
        LOG.info("Wait for pod to be Completed...")
        test_pod.wait_phase(['Succeeded', 'Failed'])
        return test_pod.read().status.phase

    def _is_k8s_obj_exist(self, k8s_res: K8sNamespacedResource):
        try:
            k8s_res.read()
        except ApiException as ex:
            try:
                body = json.loads(ex.body)
            except TypeError:
                body = {}
            if str(ex.status) == "404" and \
                    ex.reason == "Not Found" and \
                    body.get('reason') == "NotFound" and \
                    str(body.get('code')) == "404":
                LOG.info(f"Resource {k8s_res} is not found.")
                return False
        LOG.info(f"Resource {k8s_res} is present.")
        return True

    def _is_k8s_obj_not_exist(self, k8s_res: K8sNamespacedResource):
        return not self._is_k8s_obj_exist(k8s_res)

    @staticmethod
    def _check_health_status(k8s_resource: K8sNamespacedResource,
                             expected_status):
        return k8s_resource.read().status.get("health") == expected_status

    @staticmethod
    def _is_pod_restarted(pod: K8sPod, init_restarts: int):
        return pod.get_restarts_number() != init_restarts

    def wait_pod_restarted(self, pod: K8sPod, timeout=180, interval=5):
        init_restarts = pod.get_restarts_number()
        waiters.wait(
            self._is_pod_restarted,
            predicate_args=[pod, init_restarts],
            interval=interval,
            timeout=timeout,
            timeout_msg=f"K8s {pod.resource_type} {pod.name} wasn't restarted",
        )
        LOG.info(f"Number of container restarts for {pod.resource_type}"
                 f" {pod.name} was changed from {init_restarts} to"
                 f" {pod.get_restarts_number()}")

    def wait_tfoperator_healthy(self, expected_status="Ready", timeout=900,
                                interval=10):
        LOG.info("Wait until TF operator is healthy:")
        waiters.wait(
            self._check_health_status,
            predicate_args=[self.tfoperator(), expected_status],
            interval=interval,
            timeout=timeout,
            timeout_msg=f"TFOperator doesn't reached status "
                        f"`{expected_status}`",
        )
        LOG.info(f"TFOperator health status is {expected_status}.")

    def wait_tf_controllers_healthy(self, expected_status="Ready", timeout=180,
                                    interval=10):
        # TODO: implement a check of component is enabled and extend components
        #  when LCM API v2 is implemented (PRODX-17429)
        LOG.info("Wait TF controllers are healthy:")
        # Main components, enabled by default:
        resources = [self.tfconfig(), self.tfcontrol(),  self.tfvrouter()]
        if not self.is_analytics_disabled():
            resources.append(self.tfanalytics())
        for resource in resources:
            waiters.wait(
                self._check_health_status,
                predicate_args=[resource, expected_status],
                interval=interval,
                timeout=timeout,
                timeout_msg=f"TF component {resource.resource_type} "
                            f"{resource.name} doesn't reached status "
                            f"`{expected_status}`",
            )
            LOG.info(f"TF controller {resource.resource_type} {resource.name} "
                     f"is healthy.")

    def _are_pods_on_node_ready(self, node_name):
        pods = self.api.pods.list_raw(
            field_selector=f"spec.nodeName={node_name}")
        inactive = [p.metadata.name for p in pods.items if
                    p.status.phase.lower()
                    not in ('running', 'succeeded')]
        if inactive:
            LOG.info(f"Following pods are not ready: {inactive}")
            return False
        return True

    def wait_all_pods_on_node(self, node_name, timeout=300, interval=5):
        LOG.info(f"Wait all pods on node {node_name} are ready.")
        waiters.wait(
            self._are_pods_on_node_ready,
            predicate_args=[node_name],
            interval=interval,
            timeout=timeout,
            timeout_msg="Some pods aren't in running/succeeded phase",
        )

    def get_gr_job(self, read=False):
        return self.get_k8s_resource(
            "jobs", "gr-settings", self.tf_namespace, read
        )

    def wait_gr_gob(self, timeout=120, interval=3):
        job = self.get_gr_job()
        LOG.info(f"Wait {job} is succeeded.")
        waiters.wait(
            self._is_k8s_obj_exist,
            predicate_args=[job],
            interval=interval,
            timeout=timeout,
        )
        return job

    def exec_pod_cmd(self, node_name, cmd, verbose=True, timeout=600,
                     delete_pod=True):
        """Execute cmd in a privileged pod on the given node

        Doesn't require SSH access to the Node.

        * The {cmd} is executed in the privileged container from
          the chroot to the node's root filesystem /

        * A separated pod is started and removed for each {cmd}, so
          it will spend some time to pull the image and start the pod.

        * There is no way to interact with {cmd} from shell while it
          is executed. All required parameters must be set in {cmd}

        * There is no way to separate stdout from stderr, all the
          info is collected from the pod "logs".

        :rtype dict: {'logs': <str>, 'events': <str>, 'exit_code': <int>}
        """
        pod_name = "exec-pod-cmd-{}".format(utils.gen_random_string(6))
        pod_namespace = 'default'
        render_options = {
            'POD_NAME': pod_name.lower(),
            'POD_CMD': cmd,
            'NODE_NAME': str(node_name).lower(),
            'MCP_DOCKER_REGISTRY': "mirantis.azurecr.io",
        }
        pod_template = yaml.safe_load(template_utils.render_template(
            settings.MACHINE_PRIVELEGED_POD_YAML, options=render_options,
            log_env_vars=False, log_template=False))

        pod = self.api.pods.create(name=pod_name, namespace=pod_namespace,
                                   body=pod_template, log_body=False)
        if verbose:
            _LOG = LOG.info
            _ERRLOG = LOG.error
        else:
            _LOG = LOG.debug
            _ERRLOG = LOG.debug
        try:
            pod.wait_phase(['Succeeded', 'Failed'],
                           timeout=timeout, interval=1)
            pod_status = pod.read().status
            container_state = pod_status.container_statuses[0].state
            if container_state.terminated is not None:
                exit_code = container_state.terminated.exit_code
            else:
                exit_code = -1
                LOG.debug(f"Pod {pod.name} has wrong container status:\n"
                          f"{pod_status}")

            logs = pod.get_logs()
            events = None
            _LOG(f"[{node_name}] Output of the command "
                 f"\"{cmd}\":\n{logs}")
            _LOG(f"Events after command execution:\n{events}")
            _LOG(f"Exit code: {exit_code}")
            return {'logs': logs,
                    'events': events,
                    'exit_code': exit_code}
        except Exception as e:
            _ERRLOG(f"Error while running pod {pod_name}:\n{e}")
            raise (e)
        finally:
            if delete_pod:
                pod.delete()
